#### SIMULATION FUNCTIONS ####

from tkinter import *
import random
import copy

EMPTY = 0
USER = "X"
COMP = "O"

def updateView(data, canvas):
    canvas.delete(ALL)
    makeView(data, canvas)
    canvas.update()

def keyEventHandler(data, canvas, event):
    keyPressed(data, event)
    updateView(data, canvas)

def mouseEventHandler(data, canvas, event):
    mousePressed(data, event)
    updateView(data, canvas)

def runSimulation(w, h):
    data = { }
    makeModel(data)
    root = Tk()
    root.resizable(width=False, height=False) # prevents resizing window
    # We need two canvases - one for the user, one for the computer
    Label(root, text = "Tic Tac Toe").pack()
    canvas = Canvas(root, width=w, height=h)
    canvas.configure(bd=0, highlightthickness=0)
    canvas.pack()
    makeView(data, canvas)
    root.bind("<Key>", lambda event : keyEventHandler(data, canvas, event))
    root.bind("<Key>", lambda event : keyEventHandler(data, canvas, event))
    canvas.bind("<Button-1>", lambda event : mouseEventHandler(data, canvas, event))
    updateView(data, canvas)
    root.mainloop()

''' makeModel manages the state of the board. '''
def makeModel(data):
    data["board"] = [[EMPTY,EMPTY,EMPTY],[EMPTY,EMPTY,EMPTY],[EMPTY,EMPTY,EMPTY]]
    data["winner"] = None
    data["cellsize"] = 200

''' makeView manages all graphics. '''
def makeView(data, canvas):
    drawBoard(data,canvas)
    drawMoves(data,canvas)

    if data["winner"] != None:
        if data["winner"] == EMPTY:
            message = "DRAW!"
        elif data["winner"] == USER:
            message = "WIN!"
        elif data["winner"] == COMP:
            message = "LOSE!"
        canvas.create_text(300,200,font="Times 136 bold",fill="red",text=message)
        canvas.create_text(300,300,font="Times 60",text="Play Again? \n Press Enter.",fill="red",anchor="n")

''' keyPressed listens for keyboard events. '''
def keyPressed(data, event):
    if event.keysym == "Return" and data["winner"] != None:
        makeModel(data)

''' mousePressed listens for mouse events. '''
def mousePressed(data, event):
    [row, col] = getClickedCell(data, event)
    if data["winner"] != None:
        return
    runGameTurn(data, row, col)

#### BASIC FUNCTIONS ####


def checkWin(board):
    #check horizontal
    for row in board:
        if row[0] == row[1] and row[0] == row[2] and row[0] != EMPTY:
            return row[0]
    #check vertical
    for col in range(len(board[0])):
        if board[0][col] == board[1][col] and board[0][col] == board[2][col] and board[0][col] != EMPTY:
            return board[0][col]
    #check diagonal
    if board[0][0] == board[1][1] and board[0][0] == board[2][2] and board[0][0] != EMPTY:
        return board[0][0]
    if board[0][2] == board[1][1] and board[0][2] == board[2][0] and board[0][2] != EMPTY:
        return board[0][2]
    return False


def checkDraw(board):
    for row in range(len(board)):
        for col in range(len(board[row])):
            if board[row][col] == EMPTY:
                return False
    return True

def getAllNextMoves(board):
    all = []
    if checkWin(board) != False or checkDraw(board):
        return all
    for row in range(len(board)):
        for col in range(len(board[row])):
            if board[row][col] == EMPTY:
                all.append((row,col))
    return all

def drawMoves(data, canvas):
    cell = data["cellsize"]
    for row in range(len(data["board"])):
        for col in range(len(data["board"][row])):
            if data["board"][row][col] != EMPTY:
                canvas.create_text((col*cell)+(cell/2),(row*cell)+(cell/2),font="Times 136 bold",fill="white",text=data["board"][row][col])

def drawBoard(data, canvas):
    cell = data["cellsize"]
    for row in range(len(data["board"])):
        for col in range(len(data["board"][row])):
            left = col*cell
            top = row*cell
            canvas.create_rectangle(left, top, left + cell, top + cell, fill = "blue", outline="white",width="20")


def getClickedCell(data, event):
    cell = data["cellsize"]
    for row in range(len(data["board"])):
        top = row * cell
        if top <= event.y <= top + cell:
            for col in range(len(data["board"][row])):
                left = col * cell
                if left <= event.x <= left + cell:
                    return [row, col]

def checkDone(data):
    win = checkWin(data["board"])
    if win != False:
        data["winner"] = win
        return True
    draw = checkDraw(data["board"])
    if draw:
        data["winner"] = EMPTY
        return True
    return False

def runGameTurn(data, row, col):
    # If the user clicks a spot they've clicked before, don't count it
    if data["board"][row][col] == EMPTY:
        data["board"][row][col] = USER
        if not checkDone(data):
            (r,c) = getComputerAction(data)
            data["board"][r][c] = COMP
            checkDone(data)



########### EDIT THIS FUNCTION AND BELOW TO CALL RUN MINIMAX ###################
############### REMEMBER COMP IS THE PLAYER TO MAXIMIZE ###################
### HELP MAKE THE COMPUTER A SMARTER MAXIMIZER!! ###

# IMPLEMENT (3d):  higher level minimax using abmin and abmax functions
# HINT:  use abmax and/or abmin for this funciton
def getComputerAction(data):
    state = copy.deepcopy(data["board"])
    """YOUR CODE HERE"""
    return chooseFirstAvailable(data)

# this is a very simple move chooser already implemented for you
def chooseFirstAvailable(data):
    #select the first empty space
    #assumes we don't call this unless we have already checked for a draw
    return getAllNextMoves(data["board"])[0]

#should return the value and the corresponding argmin action
# this is already implemented for you
def abmin(state,alpha,beta):
    moves = getAllNextMoves(state)
    if len(moves) == 0:
        return (-1*evaluationFunction(state),None)
    v = None
    bestmove = None
    for move in moves:
        new_state = copy.deepcopy(state)
        row,col = move
        new_state[row][col] = USER
        (result,m) = abmax(new_state,alpha,beta)
        if v == None or result < v:
            v = result
            bestmove = move
        if alpha != None and v <= alpha:
            return (v,bestmove)
        if beta == None:
            beta = v
        else:
            beta = min(v,beta)
    return (v,bestmove)

# should return the value and the corresponding argmax action
# IMPLEMENT (3c): get value assuming node is max node
def abmax(state, alpha, beta):
    moves = getAllNextMoves(state)
    """YOUR CODE HERE"""
    return (0, moves[0])

# IMPLEMENT (3f): get value assuming cannot search further
# Make the computer smarter!
# HINT:  you can use functions from above such as checkWin and getAllNextMoves
def evaluationFunction(board):
    """YOUR CODE HERE"""
    return 0

runSimulation(600, 600)
